# frozen_string_literal: trues

module Hermes
  # HousingIntention model
  class HousingApplication
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::Status
    include Mongoid::Attributes::Dynamic
    
    include Hermes::Concerns::SopraXmlSender
    field :applicated_at, type: DateTime
    field :effort_rate, type: Float
    field :epci, type: Integer
    field :is_active, type: Boolean, default: true
    field :max_date, type: DateTime
    field :visit_conducted_date, type: DateTime
    field :last_status_update_date, type: Date
    field :lessor_event_informations, type: String
    field :sopra_code, type: String
    field :rank, type: Integer
    
    common_indexes *%i[
      visit_status
      visit_status_id
      visit_status_updated_at
    ]
    
    belongs_to :housing_offer, index: true, class_name: 'Hermes::HousingOffer'
    belongs_to :housing_request, index: true, class_name: 'Hermes::HousingRequest'
    
    validates_presence_of :housing_request_id, :housing_offer, unless: -> {disabled_or_deleted_status?}
    validates :status, inclusion: {in: Hermes::Statuses::HousingApplication.list[:commission], allow_blank: true}
    validates :visit_status, inclusion: {in: Hermes::Statuses::HousingApplication.list[:visit], allow_blank: true}
    
    writer_auth_roles "collaborateur offre logement", 'superadministrateur'
    writer_roles 'Superadmin', 'Administrator'
    reader_roles 'Manager'
    owner_field :user_id
    
    index(applicated_at: 1)
    index(effort_rate: 1)
    index(epci: 1)
    index(is_active: 1)
    index(max_date: 1)
    
    before_validation :before_validation_actions
    before_save :before_save_actions
    after_save :after_save_actions
    after_create :after_create_actions
    after_update :after_actions
    after_destroy :after_destroy_actions
    
    scope :selected_by_offer, ->(housing_offer_id) {
      where(housing_offer_id: housing_offer_id, status: "TRANSMIS BAILLEUR")
    }
    
    scope :actives, ->() {where(is_active: true)}
    
    HousingApplication.add_status(:single)
    HousingApplication.add_status(:single, "visit_status")

    ###############################
    #### Notification #############
    include ::Concerns::NotificationCenter::Notifiable

    def self.targets_selector(housing_app, opt = {})
      [housing_app.applicant].compact
    end

    def notification_mailer
      ::Hermes::NotificationCenter::HermesMailer
    end

    def notifiable_title(key, opt={})
      I18n.t("#{key}.title")
    end

    def notifiable_body(key, opt={})
      I18n.t("#{key}.body", number: housing_offer.reference, fullname: applicant.full_name)
    end

    def notifiable_path
      "https://als-hermes-int.be-idys.com/housing_applications/#{id}"
    end

    ###############################
    ###############################
    
    def to_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def as_json(options = {})
      options[:except] ||= :sopra_code
      super
    end
    
    def applicant
      housing_request&.main_applicant
    end
    
    def applicant_id
      applicant&.id
    end
    
    def applicated_nb_do(dec_or_inc_sym)
      return unless housing_offer
      
      case dec_or_inc_sym
      when :dec
        housing_offer.inc_applicated_nb(-1)
      when :inc
        housing_offer.inc_applicated_nb(1)
      end
    end
    
    def handle_applicated_nb
      if is_active
        applicated_nb_do :inc
      else
        applicated_nb_do :dec
      end
    end
    
    def handle_status_changed
      applicant_cancel_application if disabled_or_deleted_status?
      
      if status&.upcase == 'ACCEPTEE EN CAL'
        applicant.completed_requests_nb ? (applicant.completed_requests_nb += 1) : (applicant.completed_requests_nb = 1)
        applicant.save
        housing_request.update_applicant_eligbility
      end
      
      code_to_send = nil
      if sopra_code_changed?
        code_to_send = sopra_code
      end
      if visit_status_changed?
        case visit_status
        when 'VISITE PROPOSEE'
          code_to_send = "VISPRO"
        when 'VISITE EFFECTUEE'
          code_to_send = "VISEFF"
        end
      end
      
      return unless code_to_send
      so =  Hermes::BoClient::UpdateStatusTask.new({
                                                     housing_application_id: id,
                                                     housing_request_id: housing_request_id,
                                                     housing_offer_id: housing_offer_id,
                                                     code_evenement: code_to_send,
                                                     date_evenement: last_status_update_date || Date.current
                                                   })
      if Rails.env.test?
        {exectuted: true, xml: to_xml}
      else
       so.execute
      end
    end
    
    def handle_visit_conducted_date
      if visit_conducted_date
        if visit_conducted_date.past?
          self.set(visit_status: 'Visite effectuée') if visit_status.blank?
        else
          self.set(visit_status: 'Visite proposée') if visit_status.blank?
        end
      end
    end

    def before_save_actions
      change_is_active if status_changed?
      true
    end

    def change_is_active
      self.is_active = false if ['ANNULEE', 'ANNULEE BAILLEUR'].include?(status)
    end 
    
    def before_validation_actions
      if status_changed?
        self.status_updated_at = DateTime.current
        self.sopra_code = Hermes::Statuses::HousingApplication.sopra_code(status)
        #  self.status = Hermes::Statuses::HousingApplication.correspondence(status || '')
      end
      
      if visit_status_changed?
        self.visit_status_updated_at = DateTime.current
        self.visit_status = I18n.transliterate(visit_status || '').upcase
      end
      update_right
    end
    
    def update_right
      lessor_user_id = housing_offer&.lessor&.user_id&.to_s
      main_applicant = housing_request&.main_applicant
      reader_inst = [self, housing_request, main_applicant]
      if lessor_user_id && main_applicant
        if disabled_or_deleted_status?
          reader_inst.each {|inst| inst.readers.delete(lessor_user_id)}
          housing_request.applicant_ids.each do |co_id|
            coapplicant = Hermes::Applicant.find co_id
            coapplicant.readers.delete(lessor_user_id)
            coapplicant.set(readers: coapplicant.readers)
          end
        else
          reader_inst.each {|inst| inst.readers.push(lessor_user_id)}
          housing_request.applicant_ids.each do |co_id|
            coapplicant = Hermes::Applicant.find co_id
            if coapplicant
              coapplicant.readers.push(lessor_user_id)
              coapplicant.readers.uniq!
              coapplicant.set(readers: coapplicant.readers)
            end
          end
        end
        reader_inst.each do |inst|
          inst.readers.uniq!
          inst.set(readers: main_applicant.readers)
        end
      end
    end
    
    def array_of_selectors_to_lock
      array_of_selectors = []
      array_of_selectors.push(Hermes::HousingRequest.where(id: self.housing_request_id))
      array_of_selectors.push(Hermes::Applicant.where(id: {"$in" => [housing_request.applicant_ids, housing_request.main_applicant_id].flatten.uniq.compact}))
      array_of_selectors.push(Hermes::HousingRequestFolder.where(id: housing_request.housing_request_folder_id))
      housing_request_folder = housing_request.housing_request_folder
      array_of_selectors.push(Hermes::RequiredDocument.where(id: {"$in" => housing_request_folder.required_document_ids}))
      array_of_selectors
    
    end
    
    def lock_update_for_request
      array_of_selectors_to_lock.each do |sel|
        sel.set(is_locked_with_application: true)
      end
    end
    
    def unlock_update_for_request
      array_of_selectors_to_lock.each do |sel|
        sel.set(is_locked_with_application: false)
      end
    end
    
    def only_one_active_application
      active_applications = Hermes::HousingApplication.where({
                                                               :is_active => true,
                                                               :housing_offer_id => housing_offer_id,
                                                               :housing_request_id => housing_request_id,
                                                               :id.ne => id
                                                             })
      
      if active_applications.count
        active_applications.each do |a|
          a.is_active = false
          a.save
        end
      end
    end
    
    def update_request_active_application_id
      housing_request.active_application_id = id if is_active
      housing_request.active_application_id = nil if !is_active && housing_request.active_application_id == id.to_s
      housing_request.save if housing_request.active_application_id_changed?
    end
    
    def after_save_actions
      # Broken tests
      # check_status_changed if visit_status_changed? || status_changed?
      
      housing_request&.with_lock(reload: false) do
        housing_request.save
      end
      only_one_active_application if is_active
      update_request_active_application_id
      notify_applicant
    end
    
    def after_actions
      handle_visit_conducted_date
      handle_applicated_nb if is_active_changed?
      handle_status_changed if visit_status_changed? || status_changed?
    end
    
    def after_create_actions
      lock_update_for_request
      applicated_nb_do :inc if is_active
      housing_request.set(last_application_date: created_at, updated_at: DateTime.current)
      
      if Rails.env == 'test'
        {exectuted: true, xml: to_xml}
      else
        Hermes::BoClient::ApplicationTask.new(
          {
            housing_application_id: id,
            housing_request_id: housing_request_id,
            housing_offer_id: housing_offer_id,
          }
        ).execute
      end
    end
    
    def after_destroy_actions
      unlock_update_for_request
      housing_request.with_lock(reload: false) do
        housing_request.save
      end
      applicated_nb_do :dec
    end
    
    def disabled_or_deleted_status?
      status&.upcase&.include?('ANNULEE') || status&.upcase&.include?('DESISTEMENT')
    end
    
    def applicant_cancel_application
      if disabled_or_deleted_status? && visit_status&.upcase == 'VISITE EFFECTUEE'
        applicant.canceled_application_nb += 1
        applicant.save
        housing_request.update_applicant_eligbility
        self.set(is_active: false)
        notify(key: NOTIFICATION_KEYS[:housing_application][:applicant_cancelled_after_visit], channels: [])
      end
      housing_request.save
    end
    
    def system_cancel_application!
      set(status: 'ANNULEE', is_active: false)
      housing_request.update_applicant_eligbility
      housing_request.cancel_request(self)
    end

    def notify_applicant
      if status_changed?
        case status.upcase
          when 'ACCEPTEE EN CAL'
            # notify lessor
            # notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted])
          when 'LOGE'
            notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_accepted], channels: ['email'])
          when 'TRANSMIS BAILLEUR'
            notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_transferred], channels: ['email'])
          when 'ANNULEE BAILLEUR'
            notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_cancelled], channels: ['email'])
          when 'REFUSEE EN CAL' # status REFUSEE BAILLEUR doesent exist !!!
            notify(key: NOTIFICATION_KEYS[:housing_application][:lessor_denied], channels: ['email'])
            notify(key: NOTIFICATION_KEYS[:housing_application][:cal_denied], channels: ['email'])
          when 'EN ATTENTE'
            notify(key: NOTIFICATION_KEYS[:housing_application][:pending], channels: [])
          when 'INSCRIPTION CAL'
            notify(key: NOTIFICATION_KEYS[:housing_application][:subscribed], channels: [])
          when 'DESISTEMENT SALARIE AVANT CAL', 'DESISTEMENT APRES CAL', 'DESISTEMENT'
            notify(key: NOTIFICATION_KEYS[:housing_application][:applicant_cancelled], channels: [])
          else
            nil
        end
      end
    end
  end
end
